/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/common/ScopeTokenizer.java,v 1.2 2000/09/26 02:44:05 remm Exp $
 * $Revision: 1.2 $
 * $Date: 2000/09/26 02:44:05 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.common;

import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * Tokenizes a scope.
 * <p/>
 * This tokenizer derives from StringTokenizer and tokenizes a scope in sub scopes.
 * This object is used for Scope matching in the registry. The API of this object
 * is the same as the StringTokenizer API.
 * <p/>
 * Ex : the scope /foo/bar/bar.txt will be tokenized in the following tokens :
 * <li>/foo/bar/bar.txt
 * <li>/foo/bar
 * <li>/foo
 * <li>/
 * 
 * @see StringTokenizer
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.2 $
 */
public final class ScopeTokenizer extends StringTokenizer {
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Constructor.
     * 
     * @param uri Uri which needs to be tokenized in Scopes.
     */
    public ScopeTokenizer(Namespace namespace, String uri) {
        super(uri, "/");
        
        // We then build the scopes.
        scopes = new Vector();
        
        // Initializing namespace
        this.namespace = namespace;
        
        String courScope = new String();
        
        scopes.insertElementAt("/", 0);
        
        // We parse the token list and build the scope String objects.
        while(super.hasMoreTokens()) {
            courScope = courScope + "/" + super.nextToken();
            scopes.insertElementAt(courScope, 0);
        }
        
        pos = 0;
    }
    
    
    /**
     * Constructor.
     * 
     * @param namespace 
     * @param scope 
     */
    public ScopeTokenizer(Namespace namespace, Scope scope) {
        this(namespace, scope.toString());
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Tokens container.
     */
    Vector scopes;
    
    
    /**
     * Cursor position in the token container.
     */
    int pos;
    
    
    /**
     * Namespace.
     */
    Namespace namespace;
    
    
    // --------------------------------------------------------- Public Methods
    
    
    /**
     * Get the parent Uri of the top level scope.
     * 
     * @return Uri Parent Uri
     */
    public Uri getParentUri() {
        if (scopes.size() <= 1) {
            return null;
        } else {
            return namespace.getUri((String) scopes.elementAt(1));
        }
    }
    
    
    /**
     * True if the ScopeTokenizer contains additional scopes.
     * 
     * @return boolean
     */
    public boolean hasMoreElements() {
        return scopes.size() > pos;
    }
    
    
    /**
     * True if the ScopeTokenizer contains additional scopes.
     * 
     * @return boolean
     */
    public boolean hasMoreTokens() {
        return hasMoreElements();
    }
    
    
    /**
     * Returns the next Scope as an Object.
     * 
     * @return Object
     * @exception NoSuchElementException
     */
    public Object nextElement() 
	throws NoSuchElementException {
	Object obj = null;
	try {
	    if (!hasMoreElements()) {
		throw new NoSuchElementException();
	    }
	    obj = scopes.elementAt(pos);
	    pos = pos + 1;
	} catch (ArrayIndexOutOfBoundsException e) {
	    // Should NEVER happen.
	    e.printStackTrace();
	}
	return obj;
    }
    
    
    /**
     * Returns the next Scope as a String object.
     * 
     * @return String
     * @exception NoSuchElementException
     */
    public String nextToken()
	throws NoSuchElementException {
	return (String) nextElement();
    }
    
    
    /**
     * Returns the next Scope as a Scope object.
     * 
     * @return Scope
     * @exception NoSuchElementException
     */
    public Scope nextScope()
	throws NoSuchElementException {
	return new Scope(nextToken());
    }
    
    
    /**
     * Returns the next Scope as a String object.
     * 
     * @param delim
     * @exception NoSuchElementException
     */
    public String nextToken(String delim) 
	throws NoSuchElementException {
	return nextToken();
    }
    
    
    /**
     * Returns the scopes.
     * 
     * @return Enumeration 
     */
    public Enumeration elements() {
	return scopes.elements();
    }
    
    
    // -------------------------------------------------------- Package Methods
    
    
    /**
     * Get the parsed uri.
     * 
     * @return String Uri
     */
    String getUri() {
        if (scopes.size() < 1) {
            return null;
        } else {
            return (String) scopes.elementAt(0);
        }
    }
    
}
