/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/structure/ObjectNode.java,v 1.5 2001/02/14 07:40:44 remm Exp $
 * $Revision: 1.5 $
 * $Date: 2001/02/14 07:40:44 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.structure;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.Serializable;
import org.apache.slide.common.*;
import org.apache.slide.util.Messages;

/**
 * Represents any kind of object node.
 * <p/>
 * This includes actors, activities and collections.
 * 
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.5 $
 */
public abstract class ObjectNode 
    implements Serializable, Cloneable {
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Default constructor.
     */
    public ObjectNode() {
    }
    
    
    /**
     * Default constructor.
     */
    public ObjectNode(String uri) {
        this.uri = uri;
    }
    
    
    /**
     * Default constructor.
     */
    public ObjectNode(String uri, Vector children, Vector links) {
        this.uri = uri;
        this.children = children;
        this.links = links;
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Unique ressource identifier of the object.
     */
    protected String uri;
    
    
    /**
     * Vector of children's Uris.
     */
    protected Vector children = new Vector();
    
    
    /**
     * Vector of inbound links' Uris.
     */
    protected Vector links = new Vector();
    
    
    // ------------------------------------------------------------- Properties
    
    
    /**
     * Get object's unique resource identifier.
     *
     * @return String Uri
     */
    public String getUri() {
        return this.uri;
    }
    
    
    /**
     * Set object's unique identifier.
     *
     * @param uri Object Uri
     */
    void setUri(String uri) {
        this.uri = uri;
    }
    
    
    /**
     * Return this object's children
     *
     * @return Enumeration Children's uris
     * @deprecated
     */
    public Enumeration getChildren() {
        return enumerateChildren();
    }
    
    
    /**
     * Return this object's children
     *
     * @return Enumeration Children's uris
     */
    public Enumeration enumerateChildren() {
        return children.elements();
    }
    
    
    /**
     * Add a new child.
     * 
     * @param object Child
     */
    void addChild(ObjectNode child) {
        addChild(child.getUri());
    }
    
    
    /**
     * Add a new child to this object.
     *
     * @param uri Child's uri
     */
    void addChild(String uri) {
        if ((uri != null) && (!hasChild(uri))) {
            // We put a dummy object in the hashtable
            children.addElement(uri);
        }
    }
    
    
    /**
     * Remove all children.
     */
    void removeChildren() {
        children.removeAllElements();
    }
    
    
    /**
     * Remove child.
     * 
     * @param uri Child's Uri
     */
    void removeChild(String uri) {
        children.removeElement(uri);
    }
    
    
    /**
     * Test if object has the specified child.
     * 
     * @param object Child's uri
     * @return boolean true if this object has the specified child, 
     * false otherwise
     */
    public boolean hasChild(String uri) {
        return children.contains(uri);
    }
    
    
    /**
     * Test if this object has the specified child.
     *
     * @param child Child object
     * @return boolean true if this object has the specified child, 
     * false otherwise
     */
    public boolean hasChild(ObjectNode child) {
        boolean result = false;
        if (child != null) {
            result = children.contains(child.getUri());
        }
        return result;
    }
    
    
    /**
     * Test if object has children.
     * 
     * @return boolean true if this object has children, false otherwise
     */
    public boolean hasChildren() {
        return !(children.isEmpty());
    }
    
    
    /**
     * Return this object's inbound links
     *
     * @return Enumeration Inbound links uris
     */
    public Enumeration enumerateLinks() {
        return links.elements();
    }
    
    
    // --------------------------------------------------------- Object Methods
    
    
    /**
     * Equals.
     * 
     * @param obj Object to test
     * @return boolean True if the two object are equal :
     * <li>obj is of type ObjectNode and is not null</li>
     * <li>The Uris are equal</li>
     */
    public boolean equals(Object obj) {
        boolean result = false;
        if ((obj != null) && (obj instanceof ObjectNode)) {
            ObjectNode tempObject = (ObjectNode) obj;
            result = (this.uri.equals(tempObject.getUri()));
        }
        return result;
    }
    
    
    /**
     * Clone.
     * 
     * @return Object clone
     */
    public ObjectNode cloneObject() {
        ObjectNode result = null;
        
        try {
            result = (ObjectNode) super.clone();
            result.children = (Vector) this.children.clone();
            result.links = (Vector) this.links.clone();
        } catch(CloneNotSupportedException e) {
            e.printStackTrace();
        }
        
        return result;
    }
    
    
    /**
     * Copy.
     * 
     * @return Object copy
     */
    public ObjectNode copyObject() {
        ObjectNode result = null;
        
        try {
            result = (ObjectNode) super.clone();
            result.children = new Vector();
            result.links = new Vector();
        } catch(CloneNotSupportedException e) {
            e.printStackTrace();
        }
        
        return result;
    }
    
    
    /**
     * Validate an ObjectNode.
     * 
     * @param expectedUri Uri
     */
    public void validate(String expectedUri) {
        
        if (uri == null)
            throw new ObjectValidationFailedException
                (expectedUri, Messages.message
                 (ObjectNode.class.getName() + ".nullUri"));
        
        if (!uri.equals(expectedUri))
            throw new ObjectValidationFailedException
                (expectedUri, Messages.message
                 (ObjectNode.class.getName() + ".incorrectUri"));
        
        if (children == null)
            throw new ObjectValidationFailedException
                (uri, Messages.message
                 (ObjectNode.class.getName() + ".nullChildrenVector"));
        
        if (links == null)
            throw new ObjectValidationFailedException
                (uri, Messages.message
                 (ObjectNode.class.getName() + ".nullLinksVector"));
        
    }
    
    
}
